------------------------------------------------------
-- Aysel Karimova                                   --
-- SMALLSAT master thesis-PCA Jacobi                --
-- Supervisors: Kjetil Svarstad, Milica Orlandic    --
-- NTNU-EMECS-2018                                  --
-- Top level for PCA algorithm                      --
------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use ieee.std_logic_unsigned.all;
use work.covariance_pkg.all;
use work.array_pkg.all;

entity toplevel_cov_bram_jac is
  generic(
    matrix_size : integer := 100;
    BRAM_size   : integer := 10000;
    BRAM_width  : integer := 32;
    ADDR_width  : integer := 14;
    data_width  : positive:= 16;
    buffers     : positive:= 4
  );
  port(
    clock             : in  std_logic;
    reset             : in  std_logic;
    enable            : in  std_logic;
    cube_data         : in  std_logic_vector(data_width - 1 downto 0);
    valid_input_cov   : in  std_logic;
    output_select     : in  natural range 0 to matrix_size - 1;
    select_valid      : in  std_logic;
    valid_output      : out std_logic;
    eigenvalue        : out signed(31 downto 0);
    eigenvectors      : out signed(31 downto 0)
  );
end toplevel_cov_bram_jac;

architecture behav_top of toplevel_cov_bram_jac is

  type state_type is (st_upper_diagonal, st_lower_diagonal, st_idle, st_read_p, st_read_q, st_calc, st_write_row_p, st_write_row_q, st_write_column_p, st_write_column_q, st_read_outputs);
  signal state : state_type;
  
  component covariance_optimised is
    generic(
      data_width  : positive := 16; --size of one component
      depth_array : positive := 2500; --number of components in one plane, 2500 in our case
      planes      : positive := 100;  --number of planes, 100 in our case
      buffers     : positive := 4   --number of buffers
    );
    port(
      x             : in  std_logic_vector(data_width - 1 downto 0);
      clock, reset  : in  std_logic;
      enable        : in  std_logic;
      done          : out std_logic;
      valid_input   : in  std_logic;
      valid         : out std_logic_vector(buffers - 1 downto 0);
      column, row   : out output_row_column(buffers - 1 downto 0);
      y             : out output_t(buffers - 1 downto 0)
    );
  end component;
  
  component Block_RAM is
    generic(
      BRAM_size  : integer := 10000;
      BRAM_width : integer := 32;
      ADDR_width : integer := 14
    );
    port(
      clock   : in  std_logic;
      din     : in  std_logic_vector(BRAM_width-1 downto 0);
      we      : in  std_logic; -- write enable
      address : in  std_logic_vector(ADDR_width-1 downto 0);
      dout    : out std_logic_vector(BRAM_width-1 downto 0)
    );
  end component;
  
  component top_level_cycjacobi is
      generic(
      constant matrix_size : positive := 100;
      constant tolerance   : integer  := 100 --1000000
      );
      port(
      clock           : in  std_logic;
      reset           : in  std_logic;
      enable          : in  std_logic;
      --input data declarations: eigenvalue
      evalue_p_read   : in  row_column_array;
      evalue_q_read   : in  row_column_array;
      --input data for eigenvectors  
      evector_p_read  : in row_column_array;
      evector_q_read  : in row_column_array;
      --control signal declarations
      valid_input     : in  std_logic;
      iteration_done  : out std_logic;
      computation_done: out std_logic;
      --output data declarations: eigenvalue
      evalue_p_write  : out row_column_array;
      evalue_q_write  : out row_column_array;
      --output data eigenvectors
      evector_p_write : out row_column_array;
      evector_q_write : out row_column_array;
      --output indices
      p_latched       : out natural range 0 to matrix_size - 1; --old p to calculate the write data
      q_latched       : out natural range 0 to matrix_size - 1;
      p_new           : out natural range 0 to matrix_size - 1;
      q_new           : out natural range 0 to matrix_size - 1
      );
  end component;
  
--intermediate signal declarations
 signal column, row                     : output_row_column(buffers-1 downto 0);
 signal column_latched, row_latched     : output_row_column(buffers-1 downto 0);
 signal data_in_evalue, data_in_evector : std_logic_vector(BRAM_width-1 downto 0);
 signal address_evalue, address_evector : std_logic_vector(ADDR_width-1 downto 0);
 signal index, counter                  : integer;
 signal we                              : std_logic;
 signal cov_done, enable_jac            : std_logic;
 signal valid_in_jac                    : std_logic;
 signal jacobi_done, iteration_done     : std_logic;
 signal evalue_p_write, evalue_p_read   : row_column_array;
 signal evalue_q_read, evalue_q_write   : row_column_array;
 signal evector_p_read, evector_q_read  : row_column_array;
 signal evector_p_write, evector_q_write: row_column_array;
 signal p, q, p_new, q_new              : natural range 0 to matrix_size-1;
 signal p_latched, q_latched            : natural range 0 to matrix_size-1;
 signal y, y_latched                    : output_t(buffers-1 downto 0);
 signal valid, valid_latched            : std_logic_vector(buffers-1 downto 0);
 signal data_out_evalue, data_out_evector: std_logic_vector(BRAM_width-1 downto 0);


begin
  --port mappings
  cov_ins : covariance_optimised 
    port map(
      x           => cube_data,
      clock       => clock,
      reset       => reset,
      enable      => enable,
      done        => cov_done,
      valid_input => valid_input_cov,
      valid       => valid,
      column      => column,
      row         => row,
      y           => y
    );
  
  --BRAM instantiation for eigenvalue matrix
  bram_ins_evalue : Block_RAM 
    generic map(
      BRAM_size   => 10000,
      BRAM_width  => 32,
      ADDR_width  => 14
    )
    port map(
      clock   => clock,
      din     => data_in_evalue,
      we      => we,
      address => address_evalue,
      dout    => data_out_evalue 
    );

  --BRAM instantiation for eigenvector matrix
  bram_ins_evector : Block_RAM 
    generic map(
      BRAM_size   => 10000,
      BRAM_width  => 32,
      ADDR_width  => 14
    )
    port map(
      clock   => clock,
      din     => data_in_evector,
      we      => we,
      address => address_evector,
      dout    => data_out_evector 
    );
    
    
  jac_ins : top_level_cycjacobi
    generic map(
      matrix_size => 100,
      tolerance   => 100
    )
    port map(
      clock           => clock,
      reset           => reset,
      enable          => enable_jac,
      evalue_p_read   => evalue_p_read,
      evalue_q_read   => evalue_q_read,
      evector_p_read  => evector_p_read,
      evector_q_read  => evector_q_read,
      valid_input     => valid_in_jac,
      iteration_done  => iteration_done,
      computation_done=> jacobi_done,
      evalue_p_write  => evalue_p_write,
      evalue_q_write  => evalue_q_write,
      evector_p_write => evector_p_write,
      evector_q_write => evector_q_write,
      p_latched       => p_latched,
      q_latched       => q_latched,
      p_new           => p_new,
      q_new           => q_new
    );
    
  -- combinational logic 
  enable_jac   <= '1' when (state = st_calc) else '0';
  valid_in_jac <= '1' when (state = st_calc) else '0';

  --sequential logic
  process(clock, reset)
  begin
    if reset = '1' then
      state          <= st_idle;  --default state is st_idle
      index          <= 0;           
      data_in_evalue <= (others => '0');
      data_in_evector<= (others => '0'); 
      address_evalue <= (others => '0');
      address_evector<= (others => '0');
      eigenvalue     <= (others => '0');
      eigenvectors   <= (others => '0');
      valid_output   <= '0';
      we             <= '0';
      counter        <= 0;
      valid_latched  <= (others => '0');
      y_latched      <= (others => (others => '0'));
      column_latched <= (others => 0);
      row_latched    <= (others => 0);
    elsif clock'event and clock = '1' then
      if enable = '1' then
        case state is
        
          --write upper diagonal elements of the covariance matrix to memory
          when st_upper_diagonal =>
          --calculation of the corresponding address in upper diagonal
            address_evalue <= std_logic_vector(to_unsigned((row_latched(index) * matrix_size + column_latched(index)), address_evalue'length)); 
            address_evector <= std_logic_vector(to_unsigned((row_latched(index) * matrix_size + column_latched(index)), address_evector'length));
          --getting the data into the memory
            data_in_evalue <= std_logic_vector(y_latched(index));         --creating an identity matrix for eigenvectors
            if (row_latched(index) = column_latched(index)) then
              data_in_evector <= "00000000000000000010011100010000";
            else 
              data_in_evector <= "00000000000000000000000000000000";
            end if;
            we <= valid_latched(index);
                    
            state <= st_lower_diagonal;
            
          --duplicate the elements from upper diagonal to lower diagonal in the symmetric matrix
          when st_lower_diagonal => 
          --calculation of the corresponding address in lower diagonal
            address_evalue <= std_logic_vector(to_unsigned((column_latched(index) * matrix_size + row_latched(index)), address_evalue'length));
            address_evector <= std_logic_vector(to_unsigned((column_latched(index) * matrix_size + row_latched(index)), address_evector'length));
          --getting the data into the memory
            data_in_evalue <= std_logic_vector(y_latched(index));         
          --creating an identity matrix for eigenvectors
            if (row_latched(index) = column_latched(index)) then
              data_in_evector <= "00000000000000000010011100010000";
            else 
              data_in_evector <= "00000000000000000000000000000000";
            end if;
            we <= valid_latched(index);

            if (index < buffers-1) then
              index <= index + 1;
              state <= st_upper_diagonal;
            else 
              state <= st_idle;
            end if; 
            
          --all the output values from the covariance module are kept
--          in new values 
          --the reason is the outputs are gone after one clock cycle, 
          --thus they must be captured
          when st_idle =>
            index <= 0;
            if (unsigned(valid) > 0) then
              y_latched <= y;         --capturing the output
              valid_latched <= valid;     --capturing the valid signal 
              column_latched <= column;   --capturing the column number
              row_latched <= row;       --capturing the row number
              state <= st_upper_diagonal;   --continue with the next element
            --if covariance is finished and jacobi has not ended the computation, then
            --start reading the concerning rows. go to read row p state
            elsif ((cov_done = '1') and (jacobi_done <= '0')) then
              we <= '0';
              state <= st_read_p;
            else  
              state <= st_idle;
            end if;
              
          when st_read_p =>
          --reading row p for eigenvalue and eigenvector calc.s
            if (counter < matrix_size) then
              address_evalue <= std_logic_vector(to_unsigned(
                (p_new * matrix_size + counter), 
                address_evalue'length ));
              address_evector <= std_logic_vector(to_unsigned((p_new + counter * matrix_size), address_evalue'length ));
              counter <= counter + 1; 
              --evalue_p_read(counter) <= signed(data_out_evalue); 
            elsif (counter < matrix_size+2) then
              counter <= counter + 1;
            else
              counter <= 0;
              state <= st_read_q;
            end if;
            
            if (counter < matrix_size + 2) then
              if (counter >= 2) then
                --reading row p for eigenvector calc
                evalue_p_read(counter-2) <= signed(data_out_evalue); 
                --reading column p for eigenvector calc
                evector_p_read(counter-2) <= signed(data_out_evector);
              end if;
            end if;
                          
            
          when st_read_q =>
            if (counter < matrix_size) then
              address_evalue <= std_logic_vector(to_unsigned((q_new * matrix_size + counter), address_evalue'length ));
              address_evector <= std_logic_vector(to_unsigned((q_new + counter * matrix_size), address_evector'length ));
              counter <= counter + 1;
            elsif (counter < matrix_size + 2) then
              counter <= counter + 1;
            else
              counter <= 0;
              state <= st_calc;
            end if;
              
            if (counter < matrix_size + 2) then
              if (counter >= 2) then
                --reading row p for eigenvector calc
                evalue_q_read(counter-2) <= signed(data_out_evalue); 
                --reading column p for eigenvector calc
                evector_q_read(counter-2) <= signed(data_out_evector);    
              end if;
            end if; 
              
          when st_calc =>       --Jacobi calculates the eigenvalue 
            if (iteration_done = '1') then    --and eigenvectors
              state <= st_write_row_p;
            else 
              state <= st_calc;
            end if;
          
          when st_write_row_p =>
            if (counter < matrix_size) then
            --writes the updated row p back to bram
              address_evalue <= std_logic_vector(to_unsigned((p_latched * matrix_size + counter), address_evalue'length ));
              data_in_evalue <= std_logic_vector(evalue_p_write(counter));
              we <= '1';
              counter <= counter + 1; 
            else
              counter <= 0;
              state <= st_write_row_q;
            end if;

          when st_write_row_q =>
            if (counter < matrix_size) then
            --writes the updated row q back to bram
              address_evalue <= std_logic_vector(to_unsigned((q_latched * matrix_size + counter), address_evalue'length ));
              data_in_evalue <= std_logic_vector(evalue_q_write(counter));
              we <= '1';
              counter <= counter + 1; 
            else
              counter <= 0;
              state <= st_write_column_p;
            end if;
              
          when st_write_column_p => 
            if (counter < matrix_size) then
            --writes the updated row p back to bram
              address_evalue <= std_logic_vector(to_unsigned((p_latched + counter * matrix_size), address_evalue'length ));
              address_evector <= std_logic_vector(to_unsigned((p_latched + counter * matrix_size), address_evector'length ));
              data_in_evalue <= std_logic_vector(evalue_p_write(counter));
              data_in_evector <= std_logic_vector(evector_p_write(counter));
              we <= '1';
              counter <= counter + 1; 
            else 
              counter <= 0;
              state <= st_write_column_q;
            end if;
            
            
          when st_write_column_q =>
            if (counter < matrix_size) then
            --writes the updated row q back to bram
              address_evalue <= std_logic_vector(to_unsigned((q_latched + counter * matrix_size), address_evalue'length ));
              address_evector <= std_logic_vector(to_unsigned((q_latched + counter * matrix_size), address_evector'length ));
              we <= '1';
              data_in_evalue <= std_logic_vector(evalue_q_write(counter));
              data_in_evector <= std_logic_vector(evector_q_write(counter));
              counter <= counter + 1; 
            else
              we <= '0';
              counter <= 0;
              if (jacobi_done = '1') then
                state <= st_read_outputs;
              else
                state <= st_read_p;
              end if;
            end if;  

          --reading the results out
          when st_read_outputs => 
            if (counter < matrix_size) then
              if (select_valid = '1') then
                address_evalue <= std_logic_vector(to_unsigned((output_select * (matrix_size + 1)), address_evalue'length)); 
                address_evector <= std_logic_vector(to_unsigned((output_select + counter * matrix_size), address_evector'length));
                eigenvalue <= signed(data_out_evalue);
                eigenvectors <= signed(data_out_evector);
                --when the first element is read, set valid_output
                valid_output <= '1';
              end if;
              counter <=  counter + 1;
            elsif (counter < matrix_size + 2) then
              counter <= counter + 1;
            else
              valid_output <= '0';
              counter <= 0;
              state <= st_idle;
            end if;
            
        end case; 
      end if;
    end if;
  end process;
end behav_top;